import seqtools
import numpy
import math
import sys
import getopt
import random

from numpy import linalg

maxclasses = 16 # maximum number of rate classes allowed

b_init = 0.01
b_max = 1 # adjustment factor to get real log-det values (raise if you're 
             # getting "Very large distance" warning messages (esp. with 
             # rates-across-sites)
            

aachars = 'ARNDCEQGHILKMFPSTWYV'
aadict = {}
for i in xrange(len(aachars)):
  aadict[aachars[i]] = i
blankcounter = [0.0 for i in xrange(len(aachars))]

def remgapcols(align, rates):
  """Remove all columns with any ambiguous characters."""
  goodchars = range(len(align))
  gaps = 'x?-'

  for seq in align:
    for i in xrange(len(goodchars) - 1, -1, -1):
      if seq[goodchars[i]] in gaps:
        del goodchars[i]

  if rates:
    cleanrates = [rates[i] for i in goodchars]
  else:
    cleanrates = None
    
  cleanseqs = []
  for seq in align:
    cleanseqs.append(seq.resample(goodchars))
    cleanseqs[-1].setseqtype()

  return (seqtools.Alignment(cleanseqs), cleanrates)
  
def readrates(ratesfile):
  """Read rate classes from a file.
  
  Format is flexible. Any non-whitespace characters are assumed to be rate class
  identifiers. Importantly, each rate class must be identified by a single
  character. Whitespace is ignored completely, so identifiers can be separated
  by spaces, split over several lines, or not.
  
  """
  
  try:
    rf = open(ratesfile)
  except OSError:
    sys.stderr.write('Problem reading rate class file.\n')
    usage()
    
  rateclasses = []
  ratesdict = {}
  nclasses = 0
  whitespace = ' \t\n\x0b\x0c\r'
  
  for line in rf:
    for char in line:
      if char not in whitespace:
        try:
          rateclasses.append(ratesdict[char])
        except KeyError:
          ratesdict[char] = nclasses
          rateclasses.append(nclasses)
          nclasses += 1
          
  return (nclasses, rateclasses)

def bootstrap(align, rateclasses):
  """Bootstrap both alignment and rate class list."""
  
  nchar = len(align)
  emptyseqs = False

  while True:
  
    sites = [random.randint(0, nchar - 1) for i in xrange(nchar)]
    sites.sort()
    emptyseqs = False
  
    try:
      bootalign = resamplealign(align, sites)
    except seqtools.SeqError:
      continue
    
    if rateclasses:
      bootclasses = [rateclasses[siteID] for siteID in sites]
    else:
      bootclasses = None
  
    return (bootalign, bootclasses)

def resamplealign(align, sites):

  newseqs = []

  for seq in align:
    newseqs.append(seq.resample(sites)) 
    newseqs[-1].setseqtype()
    if newseqs[-1].type == 'empty':
      raise seqtools.SeqError('Empty sequence produced by resampling!')

  return seqtools.Alignment(newseqs)


def lddist(align, verbose=False):
  if verbose:
    print "align length:", len(align)
  dmat = []
  freqmats = []
  alphascale = calcalphascale(align)
  
  allb = []
  setb = 0

  for passnum in (1, 2):
    for i in xrange(len(align.taxlist)):
      if passnum == 2:
        dmat.append([])
      else:
        freqmats.append(buildfreqmat(align.seqlist[i]))
      for j in xrange(i):
        ldexists = True
        #print i, j, 
        #print '%s vs. %s: %g' % (align.taxlist[i], align.taxlist[j], linalg.det(builddivmat(align.seqlist[i], align.seqlist[j])))

        if passnum == 1:
          try:
            divmat = builddivmat(align.seqlist[i], align.seqlist[j])
            matdet = linalg.det(divmat)
            math.log(matdet)
            #dmat[-1].append(-math.log(matdet))
          except (ValueError, OverflowError):
            if matdet <= 0:
              if verbose:
                print 'Very large distance between %s and %s!' % (align.taxlist[i], align.taxlist[j])
              matdet, b = adjustdet(divmat)
              if b:
                if verbose:
                  print '   Fixed with a "b" adjustment coefficient of %f.' % b
                #dmat[-1].append(-math.log(matdet))
                allb.append(b)
              else:
                if verbose:
                  print '   Could not fix with a reasonable "b" coefficient.'
                else:
                  print 'Very large distance between %s and %s, setting to 2.0' % (align.taxlist[i], align.taxlist[j])
                #dmat[-1].append(2.)
                ldexists = False
            else:
              print 'Can\'t take log of determinant %g.' % matdet
              raise
          
          #if ldexists:
          #  dmat[-1][-1] += (math.log(linalg.det(numpy.dot(freqmats[i], freqmats[j]))) / 2)
          #  dmat[-1][-1] *= alphascale

        else:
          divmat = builddivmat(align.seqlist[i], align.seqlist[j])
          if setb:
            matdet, b = adjustdet(divmat, setb)
          else:
            matdet = linalg.det(divmat)
          if matdet <= 0:
            print 'Very large distance between %s and %s, setting to 2.0' % (align.taxlist[i], align.taxlist[j])
            dmat[-1].append(2.)
          else:
            dmat[-1].append(-math.log(matdet))
            dmat[-1][-1] += (math.log(linalg.det(numpy.dot(freqmats[i], freqmats[j]))) / 2)
            dmat[-1][-1] *= alphascale


    if allb and passnum == 1:
      bavg = sum(allb)/len(allb)
      stdev = 0
      for b in allb:
        stdev += (b - bavg) **2
  
      if len(allb) > 1:
        stdev = (stdev ** 0.5) /(len(allb) - 1)
      else:
        stdev = 0

      if verbose:
        print 'Average: %f; max: %f; min: %f; stdev: %f' % (bavg, max(allb), min(allb), stdev)
      
      noutlier = len(allb) / 20 # get rid of worst 5%, rounded down
      allb.sort()
      setb = allb[-noutlier]
      if verbose:
        print 'Using b adjustment factor of %f.' % setb
      
          
  return dmat

def calcalphascale(align):
  charcount = 0
  freqdict = {}
  for aa in aachars:
    freqdict[aa] = 0
    
  for seq in align:
    for char in seq.sequence:
      try:
        freqdict[char] += 1
        charcount += 1
      except KeyError:
        pass
        
  sumpisqrd = 0
  for aa in freqdict:
    sumpisqrd += (freqdict[aa]/float(charcount)) ** 2
  
  scale = (1 - sumpisqrd)/(len(aachars) - 1)
  #print 'scale:', scale
  return scale
  

def buildfreqmat(seq):
  freqmat = numpy.array([blankcounter for i in xrange(len(aachars))])

  goodcharcount = 0

  for i in xrange(len(seq)):    
    try:
      char = aadict[seq[i]]
    except KeyError: # ignore ambiguous characters
      continue
      
    goodcharcount += 1
    freqmat[char][char] += 1

  for i in xrange(len(aachars)):
    if freqmat[i][i] == 0:
      freqmat[i][i] = 0.5
    freqmat[i][i] /= float(goodcharcount)

  #print 'goodcharcount, freq mat:', goodcharcount
  return freqmat

def builddivmat(seqA, seqB, printmat=False):
  #print 'builddivmat, %s vs. %s' % (seqA.name, seqB.name)
  divmat = numpy.array([blankcounter for i in xrange(len(aachars))])

  nchar = len(seqA)
  if len(seqB) != nchar:
    sys.stderr.write('Sequences %s and %s are not the same length!\n' % (seqA.name, seqB.name))
    sys.exit(1)

  goodcharcount = 0
  for i in xrange(nchar):
    try:
      charA = aadict[seqA[i]]
      charB = aadict[seqB[i]]
    except KeyError: # ignore ambiguous characters
      continue
    
    divmat[charA][charB] += 1
    goodcharcount += 1

  if printmat:
    print divmat
    print 'goodcharcount:', goodcharcount

  for i in xrange(len(aachars)):
    for j in xrange(len(aachars)):
      if i == j and divmat[i][j] == 0:
        divmat[i][j] = 0.5
      divmat[i][j] /= float(goodcharcount)

  if printmat:
    print divmat
    print 'det:', linalg.det(divmat)
    print 'goodcharcount:', goodcharcount
  #print '%s vs. %s:' % (seqA.name, seqB.name), linalg.det(divmat)

  #print 'goodcharcount, divmat:', goodcharcount
  return divmat

def adjustdet(matrix, setb = -1):
  """Adjust the divergence matrix by Ed's formula.
  
  If F_ij is the divergence matrix, adjust by:
  
  F_ij = (1-b) F_ij + b D 
  
  Where D is the diagonal matrix of frequencies for amino acids in i and j and
  b is an arbitrary coefficient.
  
  Return det of origninal matrix and b=0 if log still can't be taken when b_max 
  is reached.
  
  """
  alphasize = len(aachars)
  
  D = numpy.array([0. for i in xrange(alphasize** 2)]).reshape(alphasize, alphasize)
  
  sum = 0
  for i in xrange(alphasize):
    for j in xrange(alphasize):
      if i == j:
        D[i,j] += matrix[i,j]
      else:
        D[i,i] += matrix[i,j]/2
        D[j,j] += matrix[i,j]/2  
        
  if setb >= 0:
    newmat = (1-setb) * matrix + setb * D
    return linalg.det(newmat), setb
        
  b = b_init
  while b <= b_max:
    newmat = (1-b) * matrix + b * D
    matdet = linalg.det(newmat)
    
    if matdet > 0:
      return (matdet, b)
    b += 0.01
    
  return (linalg.det(matrix), 0)
  
  
def summats(matrices, weights, ntaxa):

  for i in xrange(len(matrices)):
    matrices[i] = (matrices[i], weights[i])

  totalmat = []
  for i in xrange(ntaxa):
    totalmat.append([])
    for j in xrange(i):
      totalmat[-1].append(0)
      for (m, w) in matrices:
        totalmat[i][j] += m[i][j] * w
            
  return totalmat

def writephymat(matrix, taxa, outfile, lowerdiag=False, upperdiag=False):
  outfile.write(' %d\n' % len(taxa))

  for i in xrange(len(taxa)):
    outfile.write('%-10s' % taxa[i][:10])
    if lowerdiag:
      for j in xrange(i):
        outfile.write(' %9.6f' % matrix[i][j])
    elif upperdiag:
      for j in xrange(i):
        outfile.write(' ' * 10)
      for j in xrange(i + 1, len(taxa)):
        outfile.write(' %9.6f' % matrix[j][i])
        # outfile.write(' %9.6f' % matrix[i][j])
    else:
      for j in xrange(len(taxa)):
        if i < j:
          outfile.write(' %9.6f' % matrix[j][i])
        elif j < i:
          outfile.write(' %9.6f' % matrix[i][j])
        else:
          outfile.write('  0.000000')
    outfile.write('\n')
    
  outfile.flush()


def dogetopt(argv, reqdopts, legalopts):
  """Return a dictionary of options, check to make sure reqdopts are there."""
  
  opts, args = getopt.getopt(argv, legalopts)
  
  optdict = {}
  
  for (opt, val) in opts:
    optdict[opt[1:]] = val
   
  for opt in reqdopts:
    if opt not in optdict:
      raise getopt.GetoptError('Missing required option!')
      
  return (optdict, args)
  

def usage():
  sys.stderr.write('Usage: python logdet.py [options] -i seqfile -o distfile\n\n')
  sys.stderr.write('Options:\n  %-20sDo requested number of bootstraps\n' % '-b <number>')
  sys.stderr.write('  %-20sCalculate distances over separate rate classes\n' % '-r <rateClassFile>')
  sys.stderr.write('  %-20sRemove any positions with gaps before calculating distances\n' % '-c')
  sys.stderr.write('  %-20sPrint more garbage to the terminal\n' % '-v')

  sys.exit(1)

if __name__ == '__main__':


  try:
    opts, args = dogetopt(sys.argv[1:], 'io', 'i:o:cb:r:v')
  except getopt.GetoptError:
    usage()
    
  try:
    align = seqtools.getseqs(opts['i'])
    outfile = open(opts['o'], 'w')
  except OSError:
    sys.stderr.write('Problem opening input or output file.\n')
    usage()
    
  if 'b' in opts:
    try:
      nboot = int(opts['b'])
      doboot = True
    except ValueError:
      usage()
  else:
    nboot = 1
    doboot = False
      
  if 'r' in opts:
    nclasses, rates = readrates(opts['r'])
  else:
    rates = None
    nclasses = 0
    
  if nclasses > maxclasses:
    sys.stderr.write('Maximum number of rate classes is %d. Either change this')
    sys.stderr.write('value, or use fewer classes.\n')
    usage()

  if 'c' in opts:
    align, rates = remgapcols(align, rates)
  
  if 'v' in opts:
    verbose = True
  else:
    verbose = False
    
  align.seqlist = [(seq.name, seq) for seq in align.seqlist]
  align.seqlist.sort()
  align.seqlist = [seq for (name, seq) in align.seqlist]
  nchar = len(align)

  for i in xrange(nboot):
    if doboot:
      print 'Starting bootstrap %d (of %d).' % (i - 1, nboot)
      a, r = bootstrap(align, rates)
      
    else:
      a, r = align, rates

    if not nclasses:
      dmat = lddist(a, verbose)
      
    else:
      splitsites = [[] for j in xrange(nclasses)]
      for j in xrange(nchar):
        splitsites[r[j]].append(j)
      weights = [len(cls)/float(nchar) for cls in splitsites]
      #print 'weights:', weights
      
      allmats = []
      for j in xrange(nclasses):
        classalign = resamplealign(a, splitsites[j])
        allmats.append(lddist(classalign, verbose))
    
      dmat = summats(allmats, weights, len(align.taxlist))

    writephymat(dmat, align.taxlist, outfile)

  outfile.close()
  
